import json
import os

from singer import metadata
from singer.catalog import Catalog

from tap_dixa.streams import STREAMS



def get_abs_path(path):
    """
    Gets the absolute path of the provided relative path.
    """
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)

def _get_key_properties_from_meta(schema_meta):
    """
    Retrieves the 'table-key-properties' value from the schema metadata.
    """
    return schema_meta[0].get('metadata').get('table-key-properties')

def _get_replication_method_from_meta(schema_meta):
    """
    Retrieves the 'forced-replication-method' value from the schema metadata.
    """
    return schema_meta[0].get('metadata').get('forced-replication-method')

def _get_replication_key_from_meta(schema_meta):
    """
    Retrieves the 'valid-replication-keys' value from the schema metadata.
    """
    if _get_replication_method_from_meta(schema_meta) == 'INCREMENTAL':
        return schema_meta[0].get('metadata').get('valid-replication-keys')[0]
    return None

def get_schemas():
    """
    Builds the singer schema and metadata dictionaries.
    """

    schemas = {}
    schemas_metadata = {}

    for stream_name, stream_object in STREAMS.items():

        schema_path = get_abs_path('schemas/{}.json'.format(stream_name))
        with open(schema_path) as file:
            schema = json.load(file)

        if stream_object.replication_method == 'INCREMENTAL':
            replication_keys = stream_object.valid_replication_keys
        else:
            replication_keys = None

        meta = metadata.get_standard_metadata(
            schema=schema,
            key_properties=stream_object.key_properties,
            replication_method=stream_object.replication_method,
            valid_replication_keys=replication_keys,
        )

        meta = metadata.to_map(meta)

        if replication_keys:
            for replication_key in replication_keys:
                meta = metadata.write(meta, ('properties', replication_key), 'inclusion', 'automatic')

        meta = metadata.to_list(meta)

        schemas[stream_name] = schema
        schemas_metadata[stream_name] = meta

    return schemas, schemas_metadata


def discover():
    """
    Builds the singer catalog for all the streams in the schemas directory.
    """

    schemas, schemas_metadata = get_schemas()
    streams = []

    for schema_name, schema in schemas.items():
        schema_meta = schemas_metadata[schema_name]

        catalog_entry = {
            'stream': schema_name,
            'tap_stream_id': schema_name,
            'schema': schema,
            'key_properties': _get_key_properties_from_meta(schema_meta),
            'replication_method': _get_replication_method_from_meta(schema_meta),
            'replication_key': _get_replication_key_from_meta(schema_meta),
            'metadata': schema_meta
        }

        streams.append(catalog_entry)

    return Catalog.from_dict({'streams': streams})



import datetime
from typing import Iterator

import singer


def unix_ms_to_date(ms: int) -> str:
    """
    Converts unix timestamp in milliseconds to ISO 8601 date string.

    :param ms: unix timestamp in milliseconds
    :return: ISO 8601 date string
    """
    return datetime.datetime.fromtimestamp(ms/1000).replace(microsecond=0).isoformat()


def datetime_to_unix_ms(datetime_obj: datetime.datetime) -> int:
    """
    Converts datetime object to unix timestamp in milliseconds

    :param datetime_obj: A datetime object to convert to unix timestamp
    :return: integer representing unix timestamp in milliseconds
    """
    return int(datetime_obj.timestamp() * 1000)


def create_csid_params(csids: Iterator) -> dict:
    """
    Creates params for activity logs endpoint.

    Takes iterator and creates a comma-separated string list of the items
    in the iterator. Returns a dictionary with `csids` as key and
    comma-separated string list of conversation IDs as values.

    :param csids: Iterator containing conversation IDs
    :return: dictionary to be used with activity logs endpoint
    """
    return {
        'csids': ','.join(map(str, csids))
    }


def chunks(arr, chunk_size=10):
    """
    Takes an n-sized array and chunks it into smaller sizes that are
    yielded until full array has been iterated over.

    :param arr: n-sized list of conversation IDs
    :param chunk_size: the list size of each chunk
    :return: a chunk sized array
    """
    for i in range(0, len(arr), chunk_size):
        yield arr[i:i + chunk_size]


def date_to_rfc3339(date: str) -> str:
    """Converts date to rfc 3339"""
    d = singer.utils.strptime_to_utc(date)

    return d.strftime('%Y-%m-%dT%H:%M:%SZ')



import datetime
from enum import Enum

import singer
from singer import Transformer, metrics

from tap_dixa.client import Client, DixaURL
from tap_dixa.helpers import (chunks, create_csid_params, date_to_rfc3339, datetime_to_unix_ms,
                              unix_ms_to_date)

LOGGER = singer.get_logger()

class BaseStream:
    """
    A base class representing singer streams.

    :param client: The API client used extract records from the external source
    """
    tap_stream_id = None
    replication_method = None
    replication_key = None
    key_properties = []
    valid_replication_keys = []
    params = {}
    parent = None
    endpoint = None
    base_url = None

    def __init__(self, client: Client):
        self.client = client

    def get_records(self, start_date: datetime.datetime = None, is_parent: bool = False) -> list:
        """
        Returns a list of records for that stream.

        :param start_date: The start date datetime object the stream should use
        :param is_parent: If true, may change the type of data
            that is returned for a child stream to consume
        :return: list of records
        """
        raise NotImplementedError("Child classes of BaseStream require implementation")

    def set_parameters(self, params: dict) -> None:
        """
        Sets or updates the `params` attribute of a class.

        :param params: Dictionary of parameters to set or update the class with
        """
        self.params = params

    def get_parent_data(self, start_date: str = None) -> list:
        """
        Returns a list of records from the parent stream.

        :param start_date: The tap start date
        :return: A list of records
        """
        parent = self.parent(self.client)
        return parent.get_records(start_date, is_parent=True)


class Interval(Enum):
    """
    Enum representing time interval for making API calls.
    """
    HOUR = 1
    DAY = 24
    WEEK = 24 * 7
    MONTH = 24 * 31


class InvalidInterval(Exception):
    pass


class IncrementalStream(BaseStream):
    """
    A child class of a base stream used to represent streams that use the
    INCREMENTAL replication method.

    :param client: The API client used extract records from the external source
    """
    replication_method = 'INCREMENTAL'
    batched = False
    interval = None

    def __init__(self, client):
        super().__init__(client)

    def set_interval(self, value):
        """
        Sets the interval attribute.

        :param value: The interval string value
        """
        self.interval = value.upper()

    def get_interval(self):
        """
        Retrieves interval enum. Defaults to MONTH if no interval provided.
        :return: An enum for the interval to be used with the API
        """
        if self.interval:
            if hasattr(Interval, self.interval):
                return getattr(Interval, self.interval).value

            # If interval not part of enum, log message and throw error
            valid_intervals = set()
            for interval in dir(Interval):
                if not interval.startswith("__"):
                    valid_intervals.add(interval)

            # pylint: disable=logging-fstring-interpolation
            LOGGER.critical(f"provided interval '{self.interval}' is not "
                            f"in Interval set: {valid_intervals}")

            raise InvalidInterval('invalid interval provided')

        return Interval.MONTH.value

    def sync(self, state: dict, stream_schema: dict, stream_metadata: dict, config: dict, transformer: Transformer) -> dict:
        """
        The sync logic for an incremental stream.

        :param state: A dictionary representing singer state
        :param stream_schema: A dictionary containing the stream schema
        :param stream_metadata: A dictionnary containing stream metadata
        :param config: A dictionary containing tap config data
        :param transformer: A singer Transformer object
        :return: State data in the form of a dictionary
        """
        if config.get('interval'):
            self.set_interval(config.get('interval'))
        start_date = singer.get_bookmark(state, self.tap_stream_id, self.replication_key, config['start_date'])
        bookmark_datetime = singer.utils.strptime_to_utc(start_date)
        max_datetime = bookmark_datetime

        with metrics.record_counter(self.tap_stream_id) as counter:
            for record in self.get_records(bookmark_datetime):
                transformed_record = transformer.transform(record, stream_schema, stream_metadata)
                record_datetime = singer.utils.strptime_to_utc(transformed_record[self.replication_key])
                if record_datetime >= bookmark_datetime:
                    singer.write_record(self.tap_stream_id, transformed_record)
                    counter.increment()
                    max_datetime = max(record_datetime, max_datetime)

            bookmark_date = singer.utils.strftime(max_datetime)

        state = singer.write_bookmark(state, self.tap_stream_id, self.replication_key, bookmark_date)
        singer.write_state(state)
        return state


class FullTableStream(BaseStream):
    """
    A child class of a base stream used to represent streams that use the
    FULL_TABLE replication method.

    :param client: The API client used extract records from the external source
    """
    replication_method = 'FULL_TABLE'

    def __init__(self, client):
        super().__init__(client)

    def sync(self, state: dict, stream_schema: dict, stream_metadata: dict, config: dict, transformer: Transformer) -> dict:
        """
        The sync logic for an full table stream.

        :param state: A dictionary representing singer state
        :param stream_schema: A dictionary containing the stream schema
        :param stream_metadata: A dictionnary containing stream metadata
        :param config: A dictionary containing tap config data
        :param transformer: A singer Transformer object
        :return: State data in the form of a dictionary
        """
        with metrics.record_counter(self.tap_stream_id) as counter:
            for record in self.get_records(config):
                transformed_record = transformer.transform(record, stream_schema, stream_metadata)
                singer.write_record(self.tap_stream_id, transformed_record)
                counter.increment()

        singer.write_state(state)
        return state


class Conversations(IncrementalStream):
    """
    Get conversations from the Dixa platform.
    """
    tap_stream_id = 'conversations'
    key_properties = ['id']
    replication_key = 'updated_at_datestring'
    valid_replication_keys = ['updated_at_datestring']
    base_url = DixaURL.exports.value
    endpoint = '/v1/conversation_export'

    # pylint: disable=signature-differs
    def get_records(self, start_date, is_parent=False):
        created_after = start_date
        end_dt = singer.utils.now()
        add_interval = datetime.timedelta(hours=self.get_interval())
        loop = True

        while loop:
            if (created_after + add_interval) < end_dt:
                created_before = created_after + add_interval
            else:
                loop = False
                created_before = end_dt

            start = datetime_to_unix_ms(created_after)
            end = datetime_to_unix_ms(created_before)

            params = {'created_before': end, 'created_after': start}
            response = self.client.get(self.base_url, self.endpoint, params=params)

            if is_parent:
                # Chunk into max 10 csids to avoid 422 error
                # on activity logs endpoint
                conversation_ids = [record['id'] for record in response]
                yield from chunks(conversation_ids)
            else:
                for record in response:
                    record['updated_at_datestring'] = unix_ms_to_date(record['updated_at'])

                yield from response

            created_after = created_before


class Messages(IncrementalStream):
    """
    Get messages from the Dixa platform.
    """
    tap_stream_id = 'messages'
    key_properties = ['id']
    replication_key = 'updated_at_datestring'
    valid_replication_keys = ['updated_at_datestring']
    base_url = DixaURL.exports.value
    endpoint = '/v1/message_export'

    # pylint: disable=signature-differs
    def get_records(self, start_date, is_parent=False):
        created_after = start_date
        end_dt = singer.utils.now()
        add_interval = datetime.timedelta(hours=self.get_interval())
        loop = True

        while loop:
            if (created_after + add_interval) < end_dt:
                created_before = created_after + add_interval
            else:
                loop = False
                created_before = end_dt

            start = datetime_to_unix_ms(created_after)
            end = datetime_to_unix_ms(created_before)

            params = {'created_before': end, 'created_after': start}
            response = self.client.get(self.base_url, self.endpoint, params=params)

            for record in response:
                record['updated_at_datestring'] = unix_ms_to_date(end)

            yield from response

            created_after = created_before


class ActivityLogs(IncrementalStream):
    """
    Get activity logs from the Dixa platform.
    """
    tap_stream_id = 'activity_logs'
    key_properties = ['id']
    replication_key = 'activityTimestamp'
    valid_replication_keys = ['activityTimestamp']
    parent = Conversations
    base_url = DixaURL.integrations.value
    endpoint = '/v1/conversations/activitylog'

    # pylint: disable=signature-differs
    def get_records(self, start_date, is_parent=False):
        max_limit = 10_000
        total_records = max_limit
        offset = 0
        from_datetime = date_to_rfc3339(start_date.isoformat())
        to_datetime = date_to_rfc3339(datetime.datetime.utcnow().isoformat())

        params = {
            'fromDatetime': from_datetime,
            'toDatetime': to_datetime,
            'fromPage': offset,
            'limit': max_limit
        }

        while total_records == max_limit:

            response = self.client.get(self.base_url, self.endpoint, params=params)

            # Increment the offset
            params['fromPage'] += 1

            # Get the record count
            data = response.get('data', [])
            total_records = len(data)

            yield from data


STREAMS = {
    'conversations': Conversations,
    'messages': Messages,
    'activity_logs': ActivityLogs,
}



from setuptools import find_packages, setup

requirements_file = "requirements.txt"

setup(
    name="tap-dixa",
    version="0.1.0",
    description="Singer.io tap for extracting data from the Dixa API",
    author="Stitch",
    url="http://singer.io",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    py_modules=["tap_dixa"],
    install_requires=open(requirements_file).readlines(),
    entry_points="""
    [console_scripts]
    tap-dixa=tap_dixa:main
    """,
    packages=find_packages(exclude=["tests"]),
    package_data = {
        "schemas": ["tap_dixa/schemas/*.json"]
    },
    include_package_data=True,
)



import base64
from enum import Enum

import backoff
import requests
from singer import get_logger

LOGGER = get_logger()


class DixaClientError(Exception):
    def __init__(self, message=None, response=None):
        super().__init__(message)
        self.message = message
        self.response = response


class DixaClient5xxError(DixaClientError):
    pass


class DixaClient401Error(DixaClientError):
    pass


class DixaClient400Error(DixaClientError):
    pass


class DixaClient429Error(DixaClientError):
    pass


class DixaClient422Error(DixaClientError):
    pass


ERROR_CODE_EXCEPTION_MAPPING = {
    500: {
        'raise_exception': DixaClient5xxError,
        'message': 'Server Error',
    },
    503: {
        'raise_exception': DixaClient5xxError,
        'message': 'Server Error',
    },
    401: {
        'raise_exception': DixaClient401Error,
        'message': 'Invalid or missing credentials'
    },
    400: {
        'raise_exception': DixaClient400Error,
        'message': 'Invalid query parameters'
    },
    429: {
        'raise_exception': DixaClient429Error,
        'message': 'API limit has been reached'
    },
    422: {
        'raise_exception': DixaClient422Error,
        'message': 'Exceeded max allowed 10 csids per request'
    },
}


def raise_for_error(resp: requests.Response):
    """
    Raises the associated response exception.

    Takes in a response object, checks the status code, and throws the associated
    exception based on the status code.

    :param resp: requests.Response object
    """
    try:
        resp.raise_for_status()
    except (requests.HTTPError, requests.ConnectionError) as error:
        try:
            error_code = resp.status_code
            client_exception = ERROR_CODE_EXCEPTION_MAPPING.get(error_code, {})
            exc = client_exception.get('raise_exception', DixaClientError)
            message = client_exception.get('message', 'Client Error')

            raise exc(message, resp) from None

        except (ValueError, TypeError):
            raise DixaClientError(error) from None


def retry_after_wait_gen():
    """
    Returns a generator that is passed to backoff decorator to indicate how long
    to backoff for in seconds.
    """
    while True:
        sleep_time = 1
        LOGGER.info("API rate limit exceeded -- sleeping for %s seconds", sleep_time)
        yield sleep_time


class DixaURL(Enum):
    """
    Enum representing the Dixa base url API variants.
    """
    exports = 'https://exports.dixa.io'
    integrations = 'https://dev.dixa.io'


class Client:

    def __init__(self, api_token: str):
        self._api_token = api_token
        self._base_url = None
        self._session = requests.Session()
        self._headers = {}

    @staticmethod
    def _to_base64(string: str) -> str:
        """
        Base64 encodes a string.

        :param string: String to be base64 encoded
        :return: Base64 encoded string
        """
        message = f"bearer:{string}"
        message_bytes = message.encode('utf-8')
        base64_bytes = base64.b64encode(message_bytes)
        return base64_bytes.decode('utf-8')

    def _set_auth_header(self):
        """
        Sets the corresponding Authorization header based on the base url variant.
        """
        if self._base_url == DixaURL.exports.value:
            self._headers['Authorization'] = f"Basic {self._to_base64(self._api_token)}"

        if self._base_url == DixaURL.integrations.value:
            self._headers['Authorization'] = f"{self._api_token}"

    def _build_url(self, endpoint: str) -> str:
        """
        Builds the URL for the API request.

        :param endpoint: The API URI (resource)
        :return: The full API URL for the request
        """
        return f"{self._base_url}{endpoint}"

    def _get(self, url, headers=None, params=None, data=None):
        """
        Wraps the _make_request function with a 'GET' method
        """
        return self._make_request(url, method='GET', headers=headers, params=params, data=data)

    def _post(self, url, headers=None, params=None, data=None):
        """
        Wraps the _make_request function with a 'POST' method
        """
        return self._make_request(url, method='POST', headers=headers, params=params, data=data)

    @backoff.on_exception(retry_after_wait_gen, DixaClient429Error, jitter=None, max_tries=3)
    def _make_request(self, url, method, headers=None, params=None, data=None) -> dict:
        """
        Makes the API request.

        :param url: The full API url
        :param method: The API request method
        :param headers: The headers for the API request
        :param params: The querystring params passed to the API
        :param data: The data passed to the body of the request
        :return: A dictionary representing the response from the API
        """

        with self._session as session:
            response = session.request(method, url, headers=headers, params=params, data=data)

            if response.status_code != 200:
                raise_for_error(response)
                return None

            return response.json()

    def get(self, base_url, endpoint, params=None):
        """
        Takes the base_url and endpoint and builds and makes a 'GET' request
        to the API.
        """
        self._base_url = base_url
        self._set_auth_header()
        url = self._build_url(endpoint)
        return self._get(url, headers=self._headers, params=params)
